package main

import (
	"flag"
	"fmt"
	"os"
	"path/filepath"
	"sort"
	"strings"

	"git.thirdmartini.com/thirdmartini/vessel"
)

func help() {
	fmt.Printf("bundle --bundle <binary> --src <source zip>\n")
	fmt.Printf("unbundle --bundle <binary> --dst <zip content>\n")
	fmt.Printf("info --bundle <binary>\n")
	fmt.Printf("build --bundle <binary> --src <source directory>\n")
	fmt.Printf("extract --bundle <binary> --dst <source directory to extract to>\n")

	os.Exit(-1)
}

func main() {
	parms := flag.NewFlagSet("*", flag.ContinueOnError)
	bundleFlag := parms.String("bundle", "", "bundle optional zip into this file")
	srcFlag := parms.String("src", "", "source for bundle")
	dstFlag := parms.String("dst", "", "destination for bundle")
	overwriteFlag := parms.Bool("overwrite", false, "overwrite destination bundle if it exists")

	global := flag.NewFlagSet("*", flag.ContinueOnError)
	debugFlag := global.Bool("debug", false, "enable debug info")

	if len(os.Args) <= 1 {
		help()
	}

	global.Parse(os.Args)
	if *debugFlag {
		vessel.SetLogger(&vessel.DebugLogger{})
	}

	switch os.Args[1] {
	default:
		help()

	case "list":
		parms.Parse(os.Args[2:])

		b, err := vessel.Open(*bundleFlag)
		if err != nil {
			panic(fmt.Errorf("%s: %w", *bundleFlag, err))
		}
		defer b.Close()
		fmt.Printf("Bundle Size:    %d\n", b.Size())
		fmt.Printf("Bundle Version: %x\n", b.Version())
		files := b.List()
		for _, file := range files {
			fmt.Printf("  %s\n", file)
		}

	case "info":
		parms.Parse(os.Args[2:])

		b, err := vessel.Open(*bundleFlag)
		if err != nil {
			panic(fmt.Errorf("%s: %w", *bundleFlag, err))
		}
		defer b.Close()
		fmt.Printf("Bundle Size:    %d\n", b.Size())
		fmt.Printf("Bundle Version: %x\n", b.Version())
		files := b.FileInfo()

		sorted := make([]string, 0, len(files))
		for path, _ := range files {
			sorted = append(sorted, path)
		}
		sort.Strings(sorted)

		for _, path := range sorted {
			file := files[path]
			fmt.Printf("  %v  %10d  %s\n", file.ModTime().UTC(), file.Size(), path)
		}
		// print contents

	case "bundle":
		parms.Parse(os.Args[2:])
		b, err := vessel.Create(*bundleFlag, *srcFlag, *overwriteFlag)
		if err != nil {
			panic(err)
		}
		defer b.Close()
		fmt.Printf("Bundle Size:    %d\n", b.Size())
		fmt.Printf("Bundle Version: %x\n", b.Version())

	case "build":
		parms.Parse(os.Args[2:])
		b, err := vessel.Build(*bundleFlag, *srcFlag, *overwriteFlag, func(path string) bool {
			base := filepath.Base(path)
			if base == ".DS_Store" {
				return false
			}
			if strings.HasSuffix(base, ".zip") {
				return false
			}
			return true
		})
		if err != nil {
			panic(err)
		}
		defer b.Close()
		fmt.Printf("Bundle Size:    %d\n", b.Size())
		fmt.Printf("Bundle Version: %x\n", b.Version())

	case "extract":
		parms.Parse(os.Args[2:])
		b, err := vessel.Open(*bundleFlag)
		if err != nil {
			panic(err)
		}
		defer b.Close()
		fmt.Printf("Bundle Size:    %d\n", b.Size())
		fmt.Printf("Bundle Version: %x\n", b.Version())
		b.Decompress(*dstFlag)
	}
}
